#!/usr/bin/python

#This program reads a data-file, and plots the FFT of it.
#If the file has several columns, we can choose any one or the sum, or all of them.

#Copyright (C) Richard Neill 2011-2012, <maths at REMOVE.ME.richardneill.org>. This program is Free Software. You can
#redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
#either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
#For the details, please see: http://www.gnu.org/licenses/gpl.html

#TODO: print more info about the result: freq-range, peak-freq, etc   (must handle the "-c all" case too.)   [peak-freq done, add more stats?]
#TODO: add feature to dump the result to file (or stdout). Use numpy.savetxt()
#TODO: Do we need to apodise? Or would this cause its own problems with the amplitude scaling? See numpy.hanning etc.
#NOTE: the startup time is quite slow: mostly due to the "from pylab import *"

#Begin
from __future__ import division		#Needed, as this may run under Python 2.x
import getopt, sys, os, string
import numpy as numpy
from pylab import *
from textwrap import *


#Help
def usage():
	print 	"fftplot:  plot a data-file, and its frequency spectrum by FFT. Show summary information.\n"
	print 	"USAGE:    " + str(os.path.basename(sys.argv[0])) + "  [OPTS]  infile.dat\n"
	print	"OPTIONS:  -f, --samplefreq  FREQ        define the sampling frequency in Hz. [default: 1.0]."
	print   "          -c, --column      N,sum,all   use Nth column's data (N is 0-based). [default: all].\n"
	print   "          -a, --amplitude               fft y-axis: amplitude. [default: power]"
	print   "          -p, --power                   fft y-axis: power spectrum (redundant). [default: power]"
	print   "          -d, --decibels                fft y-axis: power in dB. [default: linear power]"
	print   "          -l, --logfreq                 fft x-axis: frequency on log-scale. [default: linear]\n"
        print   "          -s, --squelch_lf  FREQ        exclude frequencies <= FREQ from summary and plot. 0 hides DC."
	print   "          -z, --zoom                    zoom x-axis to \"area of interest\" (10x around modal point).\n"
	print   "          -t, --title       TITLE       override plot title"
	print   "          -x, --xlabel      LABEL       override x-axis label"
	print   "          -y, --xlabel      LABEL       override y-axis label"
	print   "          -u, --ulabel      L,A,BE,LS   override legend labels.\n"
	print   "          -r, --plotraw                 also plot the raw input file. [default: just plot FFT]."
	print   "          -o, --outplot     FILE.png    save main plot: FILE.png/.jpg, even if -n, \"auto\" names for infile.\n"
	print   "          -n, --noplot                  don't draw the plot window on screen (useful for testing/scripting)."
	print   "          -h, --help                    show help.\n"
	print   "INPUT:    Input file, infile.dat ('-' for stdin) should be columns of (int/float) data, separated by"
	print   "          whitespace, with optional #comments, i.e. the standard form for numpy.loadfile(). cf. genffttest."
	print   "          (compressed files (.gz/.bz2) can also be read transparently, no need to uncompress first.)\n"
	print   "NOTES:    * The FFT is normalised, such that a signal x = A.cos(2.PI.f.t) + B.sin(2.PI.f.t), "
	print   "          the power-spectrum has a point at (A^2 + B^2), and frequency f. But if f is unaligned,"
	print   "          i.e.  f != n*(samplerate/num_samples), then its power will be spread over several bins."
	print   "          * This does no apodisation: the window is effectively rectangular."
	print   "          * Use squelch_lf when the chart is dominated by dc/low-freq. (Can still pan to see LF)."
	print   "          * When zooming automatically (-z), the DC level (f=0) is never considered \"interesting\"."
	print   ""
	return


#Initialise
plotdims=(12, 9); plotdpi=80; 	#Sets initial size of plot on screen, and the dimensions of the saved image.  DPI affects font-size.
samplerate = plotinput = amplitude = decibels = logfreq = noplot = outplot = do_squelch_lf = title = xlabel = ylabel = ulabel = zoom = False
column = -1
squelch_lf = 0
power = True  #default anyway
the_peak_is_dc = False
info_txt = ''


#Parse args
try:
	opts, args = getopt.gnu_getopt(sys.argv[1:], "ac:df:hlnprs:o:t:x:y:u:z",
	["amplitude", "column=", "decibels", "samplefreq=", "help", "logfreq", "noplot", "power", "plotraw", "squelch_lf=", "outplot=", "title=", "xlabel=", "ylabel=", "ulabel=", "zoom"])
except getopt.GetoptError, err:
	print str(err) + ": use -h for help." ; sys.exit(2)

for o, a in opts:
	if o in ("-h", "--help"):
		usage(); exit(0);
	elif o in ("-f", "--samplefreq"):
		samplerate = float(a)
	elif o in ("-c", "--column"):
		if (a == "all") or (a == "sum"):
			column = a
		else:
			column = int(a)
	elif o in ("-r", "--plotraw"):
		plotinput = True
	elif o in ("-a", "--amplitude"):
		amplitude = True; power = False
	elif o in ("-d", "--decibels"):
		decibels = True; power = False
	elif o in ("-l", "--logfreq"):
		logfreq = True
	elif o in ("-n", "--noplot"):
		noplot = True
	elif o in ("-p", "--power"):
		power = True
	elif o in ("-o", "--outplot"):
		outplot = a
	elif o in ("-s", "--squelch_lf"):
		do_squelch_lf = True
		squelch_lf = float(a)
	elif o in ("-t", "--title"):
		title = fill(a,80)       #wrap long lines if necessary (otherwise they get truncated)
	elif o in ("-x", "--xlabel"):
		xlabel = a
	elif o in ("-y", "--ylabel"):
		ylabel = a
	elif o in ("-u", "--ulabel"):
		ulabel = a
	elif o in ("-z", "--zoom"):
		zoom = True
	else:
		assert False, "unhandled option"

if ((len(args) == 0) and ('-' in sys.argv)):		#Ugly hack for older versions.
		infile = sys.stdin
		infilename = "<stdin>"
		if (outplot == "auto"):
			outplot = "stdin.png"
elif (len(args) != 1):
		print "Error: One (non-option) argument is required: input-filename."
		exit (2)
else:
	infile = args[0]
	infilename = infile

	if (infile == "-"):
		infile = sys.stdin
		infilename = "<stdin>"
		if (outplot == "auto"):
			outplot = "stdin.png"
	else:
		if (outplot == "auto"):
			outplot = os.path.splitext(os.path.basename(infilename))[0] + ".png"


#Load data from file.
try:
	data_in = numpy.loadtxt (infile);				#numpy.loadtxt really is all we need! It even handles comments :-)
except IOError, err:							#the function is very flexible, though the defaults suffice here.
	print "Error reading datafile " + infilename + ": " + str(err)
	exit (1);
except ValueError, err:
	print "Error processing malformed datafile " + infilename + ": " + str(err)
	exit (1);


#Get size of data file, sanity check it.
if (len(data_in.shape) == 1):		#if it is a single column (1d-array),  convert to a 2d-array with 1 column.
	data_in = data_in[:,newaxis]
	column = 0;
num_cols = len(data_in[0, ]);
num_rows = len(data_in);
if (num_cols == 0) or (num_cols > 20) or (num_rows < 4) or (num_rows > 1E7):
	print "Data size is out of range. File contains "  + str(num_rows) + " rows of " + str(num_cols) + " columns."
	exit (1)


#Extract the relevant column.
if (column == -1):			#Default: All columns.
	column = "all"
	if (num_cols == 1):		#..but if the file only has 1 column, don't make it an array (for clarity).
		column = 0
if (column == "sum"):			#Sum the columns
	data = data_in [ : , 0]
	for col in range (1, num_cols):
		data += data_in [ : , col]
	legend = ["sum (0:"+str(num_cols -1)+")"]
elif (column == "all"):			#All columns (in parallel).
	data = data_in
	legend = ["col 0"]
	for col in range (1, num_cols):
		legend.append ("col " +str(col))
elif (column < num_cols):
	data = data_in [ : , column]	#This pulls out a 1-d array, from a 2-d array.
	legend = ["col "+str(column)]
else:
	assert False, "invalid column specified: " + str(column)


#Sampling number and rate.
num_samples = len(data)			#This many samples.
if (samplerate == False):		#May be given as arg, else 1.0 (float)
	print "Sample frequency (-f) not specified; assuming 1.0 Hz"
	samplerate = 1.0
times =	arange (0, num_samples) * (1.0/samplerate) #Create time-series for raw-data x-axis, instead of just showing sample-number
duration = num_samples / samplerate
binwidth = samplerate / num_samples	#Full-Width of each frequency "bin" in Hz.
if (squelch_lf < 0 or squelch_lf >= (samplerate/2.0)):
	print "Error: squelch_lf must be positive, and less than (samplerate/2)."
	exit (1);


#Do the FFT.
#The discrete fourier transform is documented here: http://docs.scipy.org/doc/numpy/reference/routines.fft.html
if (data.ndim == 1):
	fft_data = rfft(data)		#FFT, for real input data. (returns complex array). The factor (2) below is from FFT maths: scales amplitude correctly.
else:
	fft_data = rfft(data[ : ,0])	#Annoyingly, rfft() chokes on multi-column array. Iterate.
	for col in range (1, num_cols):
		tmp = rfft(data[ : ,col])
		fft_data = column_stack ((fft_data, rfft(data[ : ,col]) ))

fft_data = 2 * fft_data / num_samples	#Scale it, so it is independent of sample count. The 2 comes from FFT maths. Input "A cos (wt)" now has FFT amplitude A' = A.
fft_data[0] /= 2			#... but don't double the DC-level.
fft_abs =  abs (fft_data) 		#Convert the sin/cos complex components to absolute amplitude value.
fft_power = fft_abs **2			#Square, to get power-spectrum.
fft_power_db = 10 * log10 (fft_power);  #Convert power-spectrum to dB, if desired.

frequencies = fftfreq (num_samples, (1.0/samplerate))   #Calculate Frequency bins. (Point 0 is dc-level)
frequencies = frequencies[0: len(fft_data)]		#Now trim off the negative frequencies (from the end).
frequencies[-1] = abs(frequencies[-1]);			#Flip the final frequency-value to positive, if necessary (when num_samples is even)


#Find the peak (modal) frequency (or array). Then find the zoom-control point.
#zoom_x, zoom_y is the modal point. excepting that, if this is the dc value, x is moved to the non-dc-modal point, without changing y.
offset = 1;					#Exclude the 0th point (DC level) from this search (start searching from position 1, but answer is wrt offset=0.)
maxpos_ac = fft_abs[offset:].argmax(axis=0) + offset
offset = 1 + int(math.floor(squelch_lf / binwidth))	#... again, but excluding the low frequency points.
maxpos_ac_notlf = fft_abs[offset:].argmax(axis=0) + offset
if (data.ndim == 1):
	fft_peak_ac  = fft_abs[maxpos_ac]	#Amplitude
	f_peak_ac = frequencies[maxpos_ac]	#Corresponding Frequency
	fft_peak_ac_notlf = fft_abs[maxpos_ac_notlf]   #...excluding the squelched LF.
	f_peak_ac_notlf = frequencies[maxpos_ac_notlf]
	fft_value_dc = fft_abs[0]		#... the (obvious) DC values.
	f_value_dc = frequencies[0]

	if (do_squelch_lf):			#Get most interesting zoom point. If we're squelching LF, look at the high-freq AC only.
		zoom_x = f_peak_ac_notlf
		zoom_y = fft_peak_ac_notlf
	elif (fft_peak_ac > fft_value_dc):	#Else, look whether the AC is stronger than the DC.
		zoom_x = f_peak_ac		# ...yes, so look at the AC peak.
		zoom_y = fft_peak_ac
	else:
		zoom_x = f_peak_ac		# ...no, so look at the DC spike, and set the width by the AC-peak's position.
		zoom_y = fft_value_dc
		the_peak_is_dc = True;

else:						#Array versions of the above.
	fft_peak_ac  = diagonal(fft_abs[maxpos_ac])
	f_peak_ac = frequencies[maxpos_ac]
	fft_peak_ac_notlf = diagonal(fft_abs[maxpos_ac_notlf])
	f_peak_ac_notlf = frequencies[maxpos_ac_notlf]
	fft_value_dc = fft_abs[0]
	f_value_dc = frequencies[0]
	
	maxpos_ac_pt = fft_peak_ac.argmax(axis=0) 	#combine all channels, get the biggest.
	fft_peak_ac_pt = fft_peak_ac[maxpos_ac_pt]	#now get single point, not array
	f_peak_ac_pt = f_peak_ac[maxpos_ac_pt]
	maxpos_ac_notlf_pt = fft_peak_ac_notlf.argmax(axis=0)
	fft_peak_ac_notlf_pt = fft_peak_ac_notlf[maxpos_ac_notlf_pt]
	f_peak_ac_notlf_pt = f_peak_ac_notlf[maxpos_ac_notlf_pt]
	
	if (do_squelch_lf):
		zoom_x = f_peak_ac_notlf_pt
		zoom_y = fft_peak_ac_notlf_pt
	elif (fft_peak_ac > fft_value_dc).any():
		zoom_x = f_peak_ac_pt
		zoom_y = fft_peak_ac_pt
	else:
		zoom_x = f_peak_ac_pt;
		zoom_y = fft_value_dc[0]
		the_peak_is_dc = True;
		

#Plot the raw input data, if desired.
if (plotinput):
	plt.figure(figsize=plotdims, dpi=plotdpi);
	plt.title ("Raw signal data  [ " + infilename + ", column " + str(column) + " ]")
	plt.xlabel("Time (s)")		#Dimensional assumption that x represents time-in-seconds is justified, given the definition of -f
	plt.ylabel("Signal (V)")	#This assumption (choosing Volts, rather than mV, or photons) is purely arbitrary, but increases clarity.
	lines = plt.plot (times, data)	#If data is a multi-column array, this plots multiple lines :-)
	plt.setp (lines, marker = '.', linestyle='-');	#Plot with solid lines and small dots.
	plt.grid()
	if (num_cols > 1):
		plt.legend(legend)
	info_txt += "Raw data and "


#Plot the fft data
plt.figure(figsize=plotdims, dpi=plotdpi);
if (amplitude):				#Plot Amplitude vs freq.
	plt.title  ("FFT Amplitude  [ " + infilename + ", column " + str(column) + " ]")
	plt.xlabel ("Frequency (Hz)")
	plt.ylabel ("Amplitude (V)")
	lines = plt.plot (frequencies, fft_abs)
	info_txt += "FFT (Amplitude). "

elif (power):				#Plot power-spectrum (default)
	plt.title  ("FFT Power Spectrum  [ " + infilename + ", column " + str(column) + " ]")
	plt.xlabel ("Frequency (Hz)")
	plt.ylabel (r'Power ($V^2$)')	#embedded TeX.
	lines = plt.plot   (frequencies, fft_power)
	info_txt += "FFT Power Spectrum. "

elif (decibels):			#Plot power spectrum, in dB
	plt.title  ("FFT Power Spectrum, dB  [ " + infilename + ", column " + str(column) + " ]")
	plt.xlabel ("Frequency (Hz)")
	plt.ylabel ("Power (dB)")
	lines = plt.plot (frequencies, fft_power_db)
	info_txt += "FFT Power Spectrum (dB). "
else:
	assert False, "plot what?"

#Plot with solid lines and small dots. (Automatic colors). Important to have both, otherwise spikes get "moire interference" wrt the plotting grid, which makes
plt.setp (lines, marker = '.', linestyle='-'); #the FFT freq peaks disappear, or have much lower apparent-amplitude!
if (num_cols > 1):
	plt.legend(legend)
plt.grid ()  #add grid


#Log frequency scale:
if (logfreq):
	plt.xscale ("log")
	info_txt += ""


#Zoom in on x-axis area of greatest interest. Guess that this is around the modal point (but NB DC, see definition of zoom_x and zoom_y above).
if (zoom):
	plt.xlim ([ 0.1 * zoom_x, 10 * zoom_x])			#Arbitrary: zoom in on frequency from  f/10 to 10*f.
	if (the_peak_is_dc):					# ... unless peak is at zero, in which case, zoom from
		plt.xlim ([ -0.05 * zoom_x, 10 * zoom_x])	#       -0.05*f_ac to 10*f_ac, where f_ac is first non-zero peak
	if (10 * zoom_x > frequencies[-1]):			# ... subject to being within overall range.
		plt.xlim ([0 ,frequencies[-1] ])

	plt.xlabel ("Frequency (Hz),  cropped")			#The X-axis crop could lose data, so warn; the Y-axis crop (except dB) can't.
	
elif (do_squelch_lf):						#Squelching LF, but without automatic zoom to most interesting frequency-zone.
	plt.xlim(squelch_lf, frequencies[-1])			#X axis set to [lf_threshold, max-freq]
	plt.xlabel ("Frequency (Hz), above " + str(squelch_lf) + " Hz")


if (zoom or do_squelch_lf):					#Set the y-axis sensibly for our choice of x.
	if (amplitude):
		plt.ylim ([ -0.05 * zoom_y,  1.1 * zoom_y ]) 	#zoom in on amplitude from -Y_max/20  to 1.1*Y_max
	elif (power):
		zoom_y = zoom_y**2				#likewise (but squared)
		plt.ylim ([ -0.05 * zoom_y,  1.1 * zoom_y ])
	elif (decibels):
		zoom_y = log10 (zoom_y**2)			#dB range is 30 below to 10 above.
		plt.ylim ([zoom_y - 30, zoom_y + 10])


#Override labels, if specified.
if (title != False):
	plt.title (title)
if (xlabel != False):
	plt.xlabel (xlabel)
if (ylabel != False):
	plt.ylabel (ylabel)
if (ulabel != False):
	ulabels = string.split (ulabel, ",", (num_cols - 1) )	#comma separated.
	plt.legend (ulabels)
	if ((column < num_cols) and (len(ulabels) >= column)):  #select nth label if possible.
		plt.legend ([ulabels[column]])


#Show summary info.
print "Loaded column " + str(column) + " from file '" + infilename + "': " + str(num_samples) + " points at " + str(samplerate) + " Hz sample-rate, for " + str(duration) + " seconds."
print "Plotting " + info_txt + "Nyquist frequency is " + str(samplerate/2) + " Hz. Frequency bins are " + str(binwidth) + " Hz wide (i.e. +/- " + str(binwidth/2) + " Hz)."
print "The D.C. Amplitude is ", fft_value_dc, ", occurring at frequency ", f_value_dc , " Hz, i.e. D.C. "    	#Don't use str() here: it fails if "-c all"
print "Max A.C. Amplitude is ", fft_peak_ac, ", occurring at frequency ", f_peak_ac , " Hz."
if (do_squelch_lf):
	print "Max H.F. Amplitude is ", fft_peak_ac_notlf, ", occurring at frequency ", f_peak_ac_notlf , " Hz, for f > ", squelch_lf, " Hz."


#Now save and show the plots, and exit.
if (outplot):
	plt.savefig (outplot)
	print "Saved file as: " + outplot
if (noplot == False):
	plt.show()
exit (0)
